from .error import ValidateError
from .error import ImplementationError


class MatcherBase(object):
    def __init__(self, *args, **kwargs):
        self.init_arg_str(*args, **kwargs)
        self.initialize(*args, **kwargs)

    def __ne__(self, rhs):
        return not self.__eq__(rhs)

    def __eq__(self, rhs):
        try:
            self.validate(rhs)
            return True
        except ValidateError:
            return False

    def __str__(self):
        return "%s(%s)" % (self.__class__.__name__, self.__arg_str)

    def __repr__(self):
        return self.__str__()

    def init_arg_str(self, *args, **kwargs):
        list_of_arg = list(args)
        for k, v in kwargs.items():
            list_of_arg.append("%s=%s" % (k, v))
        self.__arg_str = ", ".join([str(i) for i in list_of_arg])

    def raise_validate_error(self, rhs, **kwargs):
        raise ValidateError(self, rhs, **kwargs)

    def validate(self, rhs, e_path=""):
        raise ImplementationError("Need implementation")

    def initialize(self, *args, **kwargs):
        raise ImplementationError("Need implementation")


def validate(v1, v2, e_path=""):
    if hasattr(v1, "validate"):
        v1.validate(v2, e_path=e_path)
    elif hasattr(v2, "validate"):
        v2.validate(v1, e_path=e_path)
    elif v1 != v2:
        raise ValidateError(v1, v2, e_path=e_path)



from .internal import MatcherBase


class Nullable(MatcherBase):
    def initialize(self, obj):
        self.obj = obj

    def validate(self, rhs, e_path=""):
        if rhs is None:
            return
        if self.obj != rhs:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_param")


class Not(MatcherBase):
    def initialize(self, obj):
        self.obj = obj

    def validate(self, rhs, e_path=""):
        if self.obj == rhs:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_param")


class Any(MatcherBase):
    def initialize(self, *type_classes):
        self.type_classes = type_classes
        self.init_arg_str(*[i.__name__ for i in type_classes])

    def validate(self, rhs, e_path=""):
        if not self.type_classes:
            return
        if not isinstance(rhs, self.type_classes):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_param")



from .internal import MatcherBase
from .internal import validate


class OneOf(MatcherBase):
    def initialize(self, *candidates):
        self.candidates = candidates

    def validate(self, rhs, e_path=""):
        for i in self.candidates:
            if i == rhs:
                return
        self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_argument")


class AnyOrder(MatcherBase):
    def initialize(self, list_obj):
        assert isinstance(list_obj, list)

        self.list_obj = list_obj

    def validate(self, rhs, e_path=""):
        if len(rhs) != len(self.list_obj):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="different_length")
        list_obj = self.list_obj[:]
        for idx, i in enumerate(rhs):
            if not i in list_obj:
                self.raise_validate_error(rhs, e_path=e_path, e_msg=f'item_not_found_at_index: {idx}')
            list_obj.remove(i)


class ListOf(MatcherBase):
    def initialize(self, one_item, n_item=None, min_item=None, max_item=None):
        self.one_item = one_item
        self.n_item = n_item
        self.min_item = min_item
        self.max_item = max_item

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, list):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_list")
        if self.n_item is not None:
            if not len(rhs) == self.n_item:
                self.raise_validate_error(rhs, e_path=e_path, e_msg=f'mismatch_item_count: {len(rhs)} != {self.n_item}')
        if self.min_item is not None:
            if len(rhs) < self.min_item:
                self.raise_validate_error(rhs, e_path=e_path, e_msg=f'too_few_items: {len(rhs)} < {self.min_item}')
        if self.max_item is not None:
            if len(rhs) > self.max_item:
                self.raise_validate_error(rhs, e_path=e_path, e_msg=f'too_many_items: {len(rhs)} > {self.max_item}')
        for idx, val in enumerate(rhs):
            validate(self.one_item, val, "%s.%s" % (e_path, str(idx)))



from .internal import MatcherBase
from . import number_matcher
from . import str_matcher
from . import list_matcher
from . import dict_matcher
from . import other_matcher


class BelieveMixin(object):
    pass


# Put all Matcher into BelieveMixin
for m in [number_matcher, str_matcher, list_matcher, dict_matcher, other_matcher]:
    for c in dir(m):
        cls = getattr(m, c)
        try:
            if issubclass(cls, MatcherBase) and cls != MatcherBase:
                setattr(BelieveMixin, c, cls)
        except TypeError:
            pass



import uuid

from .internal import MatcherBase


class AnyStr(MatcherBase):
    def initialize(self, min_len=None, max_len=None):
        self.min_len = min_len
        self.max_len = max_len

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_string")
        if self.min_len is not None and len(rhs) < self.min_len:
            self.raise_validate_error(rhs, e_path=e_path, e_msg=f'string_too_short: {len(rhs)} < {self.min_len}')
        if self.max_len is not None and len(rhs) > self.max_len:
            self.raise_validate_error(rhs, e_path=e_path, e_msg=f'string_too_long: {len(rhs)} > {self.max_len}')


class AnyIntStr(MatcherBase):
    def initialize(self):
        pass

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_string")

        try:
            int(rhs)
        except ValueError:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_int_string")


class AnyUUID(MatcherBase):
    def initialize(self):
        pass

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, str) or len(rhs) != 36:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_uuid")
        try:
            uuid.UUID(rhs)
        except (TypeError, ValueError):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_uuid")


class AnyIPV4(MatcherBase):
    def initialize(self):
        pass

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_ipv4")
        tokens = rhs.split('.')
        if len(tokens) != 4:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_ipv4")
        for token in tokens:
            try:
                token = int(token)
            except ValueError:
                self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_ipv4")
            if not 0 <= token < 256:
                self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_ipv4")


class AnySHA1(MatcherBase):
    def initialize(self):
        pass

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, str):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_sha1")
        if len(rhs) != 40:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_sha1")
        try:
            int(rhs, 16)
        except ValueError:
            self.raise_validate_error(rhs, e_path=e_path, e_msg="invalid_sha1")



from .internal import MatcherBase


class Almost(MatcherBase):
    def initialize(self, ts, ts_range=3):
        assert isinstance(ts, (int, float))
        assert isinstance(ts_range, (int, float))

        self.ts = ts
        self.ts_range = ts_range

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, (int, float)):
            self.raise_validate_error(rhs, e_path=e_path, e_msg=f'not_int_or_float')
        if not self.ts - self.ts_range <= rhs <= self.ts + self.ts_range:
            self.raise_validate_error(rhs, e_path=e_path, e_msg=f'not_in_range')


class AnyInt(MatcherBase):
    def initialize(self, min_value=None, max_value=None):
        assert min_value is None or isinstance(min_value, int)
        assert max_value is None or isinstance(max_value, int)

        self.min_value = min_value
        self.max_value = max_value

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, int):
            self.raise_validate_error(rhs, e_path=e_path)
        if self.min_value is not None and self.min_value > rhs:
            self.raise_validate_error(rhs, e_path=e_path, e_msg=f'value_too_small: {rhs} < {self.min_value}')
        if self.max_value is not None and self.max_value < rhs:
            self.raise_validate_error(rhs, e_path=e_path, e_msg=f'value_too_large: {rhs} > {self.max_value}')



from .internal import MatcherBase
from .internal import validate


class Dict(MatcherBase):
    def initialize(self, dict_obj):
        self.dict_obj = dict_obj
        assert isinstance(dict_obj, dict)

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, dict):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_dict")

        # validate required field
        for k, v in self.dict_obj.items():
            if not isinstance(v, Optional):
                if k not in rhs:
                    self.raise_validate_error(rhs, e_path=e_path, e_msg=f'missing_required_field: {k}')

        # validate field value
        for k, v in rhs.items():
            if k in self.dict_obj:
                validate(self.dict_obj[k], v, "%s.%s" % (e_path, k))
            else:
                self.raise_validate_error(rhs, e_path=e_path, e_msg="unknown_field", e_unsafe_msg=f'unknown_field: {k}')

class DictOf(MatcherBase):
    def initialize(self, key, value, n_item=None, min_item=None, max_item=None):
        self.key = key
        self.value = value
        self.n_item = n_item
        self.min_item = min_item
        self.max_item = max_item

    def validate(self, rhs, e_path=""):
        if not isinstance(rhs, dict):
            self.raise_validate_error(rhs, e_path=e_path, e_msg="not_dict")
        if self.n_item is not None:
            if not len(rhs) == self.n_item:
                self.raise_validate_error(rhs, e_path=e_path, e_msg=f'mismatch_item_count: {len(rhs)} != {self.n_item}')
        if self.min_item is not None:
            if len(rhs) < self.min_item:
                self.raise_validate_error(rhs, e_path=e_path, e_msg=f'too_few_items: {len(rhs)} < {self.min_item}')
        if self.max_item is not None:
            if len(rhs) > self.max_item:
                self.raise_validate_error(rhs, e_path=e_path, e_msg=f'too_many_items: {len(rhs)} > {self.max_item}')
        for k, v in rhs.items():
            validate(self.key, k, "%s.%s" % (e_path, k))
            validate(self.value, v, "%s.%s" % (e_path, k))

class Optional(MatcherBase):
    def initialize(self, value):
        self.value = value

    def validate(self, rhs, e_path=""):
        validate(rhs, self.value, e_path=e_path)



class ValidateError(Exception):
    def __init__(self, v1, v2, **kwargs):
        self.v1 = repr(v1)
        self.v2 = repr(v2)
        self.kwargs = kwargs

    def kwargs_to_string(self, fields, with_v2):
        result = []
        for k in fields:
            if self.kwargs.get(k):
                if k == 'e_path':
                    result.append(f'[{k}=${self.kwargs[k]}]')
                else:
                    result.append(f'[{k}={self.kwargs[k]}]')
        if with_v2:
            result.append(f'{self.v2} != {self.v1}')
        return " ".join(result)

    def xss_unsafe_message(self):
        return self.kwargs_to_string(['e_path', 'e_msg', 'e_unsafe_msg'], True)

    def xss_safe_message(self):
        return self.kwargs_to_string(['e_path', 'e_msg'], False)

    def __str__(self):
        return self.xss_unsafe_message()


class ImplementationError(Exception):
    pass



import os

from setuptools import find_packages, setup

try:
    version = os.environ['VERSION']
except:
    version = '1.0.3'

# read the contents of your README file
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, '../', 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

if __name__ == '__main__':
    setup(
        name='believe',
        version=version,
        description='A easy to use validator for json content',
        long_description=long_description,
        long_description_content_type='text/markdown',
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent"
        ],  # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
        python_requires='>=3.0',
        keywords=['json', 'validate', 'validator'],
        author='Seth Wang',
        author_email='pkyosx@gmail.com',
        url='https://github.com/pkyosx/believe',
        packages=find_packages(exclude=["ez_setup", "examples", "*.tests", "*.tests.*", "tests.*", "tests"]),
        include_package_data=True,
    )



from .mixin import BelieveMixin
from .error import ValidateError
from .error import ImplementationError

from .internal import MatcherBase
from .internal import validate

from .dict_matcher import Dict
from .dict_matcher import DictOf
from .dict_matcher import Optional

from .list_matcher import OneOf
from .list_matcher import AnyOrder
from .list_matcher import ListOf

from .number_matcher import Almost
from .number_matcher import AnyInt

from .other_matcher import Nullable
from .other_matcher import Not
from .other_matcher import Any

from .str_matcher import AnyStr
from .str_matcher import AnyIntStr
from .str_matcher import AnyUUID
from .str_matcher import AnyIPV4
from .str_matcher import AnySHA1
